/****************************************************************/
/**@file hudaq_internal.h:
 * Description: Internal hudaq header with private information. *
 * Dependency: Only for Linux                                   *
 *                Copyright 2006-2007 Humusoft s.r.o.           *
 ****************************************************************/

#ifndef HUDAQ_INTERNAL_H__
#define HUDAQ_INTERNAL_H__

#include "types.h"

#define __int8 char
#define __int16 short int
#define __int32 int

#ifndef max
 #define max(x,y) (((x)>(y))?(x):(y))
#endif
#ifndef min
 #define min(x,y) (((x)>(y))?(x):(y))
#endif


/**
 * Resource information structure obtained from low level kernel driver.
*/
typedef struct
{
  size_t Size;                        ///< Size of this structure.
  int NumPhysMemResources;            ///< Amount of memory mapped ranges in particular device.
  HudaqResourceRange PhysMemResources[8];///< Memory resources available for user access.
  HudaqResourceInfo Resources;        ///< External structure for users
  size_t DriverDataSize;              ///< Size of shared memory among all applications.
  void* DriverData;                   ///< Internal data for one device shared across all applications.

        // TODO: these two items are planned to be filled from kernel driver
} DriverRecord;


typedef struct CallTable;

/** Device record that contains common information about general device resources. */
typedef struct
{
//HANDLE SysHandle;             ///< System handle assigned from kernel driver
  char Name[256];               ///< Device name in ASCII
  int Order;                    ///< Order of device
  DriverRecord DrvRes;          ///< Kernel resource info

  const struct CallTable *pCT;  ///< Interface functions
} DeviceRecord;


typedef int (TInitDev)
   (DeviceRecord *DevRecord, int IniOptions);
typedef void (TDoneDev)
   (DeviceRecord *DevRecord);

typedef int (THudaqXXRead)
   (const DeviceRecord *DevRecord, unsigned channel);

typedef HUDAQSTATUS (THudaqDIReadMultiple)
   (const DeviceRecord *DevRecord, unsigned number, const unsigned* channels, unsigned* values);
typedef int (THudaqDIReadBit)
   (const DeviceRecord *DevRecord, unsigned channel, unsigned bit);

typedef HUDAQSTATUS (THudaqDOWrite)
   (const DeviceRecord *DevRecord, unsigned channel, unsigned value);
typedef void (THudaqDOWriteBit)
   (const DeviceRecord *DevRecord, unsigned channel, unsigned bit, int value);
typedef void (THudaqDOWriteMultipleBits)
   (const DeviceRecord *DevRecord, unsigned channel, unsigned mask, unsigned value);

typedef double (THudaqAIRead)
   (const DeviceRecord *DevRecord, unsigned channel);
typedef HUDAQSTATUS (THudaqAIReadMultiple)
   (const DeviceRecord *DevRecord, unsigned number, const unsigned *channels, double *values);

typedef void (THudaqAOWrite)
   (const DeviceRecord *DevRecord, unsigned channel, double value);
typedef HUDAQSTATUS (THudaqAOWriteMultiple)
   (const DeviceRecord *DevRecord, unsigned number, const unsigned* channels, const double* values);

typedef double (THudaqGetParameter)
   (const DeviceRecord *DevRecord, unsigned channel, HudaqParameter param);

typedef HUDAQSTATUS (THudaqSetParameter)
   (const DeviceRecord *DevRecord, unsigned channel, HudaqParameter param, double value);

typedef const HudaqRange *(THudaqQueryRange)
   (const DeviceRecord *DevRecord, HudaqSubsystem S, unsigned item);

typedef  HUDAQSTATUS (THudaqStepWrite)
   (const DeviceRecord *DevRecord, unsigned channel, int position);

typedef HUDAQSTATUS (THudaqPWMWrite)
   (const DeviceRecord *DevRecord, unsigned channel, double frequency, double dutycycle);

typedef HUDAQSTATUS (THudaqReset)
   (const DeviceRecord *DevRecord, unsigned channel);


/** Call table of HUDAQ API
  This table should not contain NULL inside any field. Please use empty caller from generic.c
  instead. */
typedef struct CallTable
{
  const char *DeviceString;

  TInitDev *HudaqInit;
  TDoneDev *HudaqDone;

  THudaqSetParameter *HudaqSetParameter;
  THudaqGetParameter *HudaqGetParameter;
  THudaqQueryRange *HudaqQueryRange;

  THudaqXXRead *HudaqDIRead;
  THudaqDIReadBit *HudaqDIReadBit;
  THudaqDIReadMultiple *HudaqDIReadMultiple;

  THudaqDOWrite *HudaqDOWrite;
  THudaqDOWriteBit *HudaqDOWriteBit;
  THudaqDOWriteMultipleBits *HudaqDOWriteMultipleBits;

  THudaqAIRead *HudaqAIRead;
  THudaqAIReadMultiple *HudaqAIReadMultiple;

  THudaqAOWrite *HudaqAOWrite;
  THudaqAOWriteMultiple *HudaqAOWriteMultiple;

  THudaqXXRead *HudaqEncRead;
  THudaqReset  *HudaqEncReset;

  THudaqXXRead *HudaqCtrRead;
  THudaqReset  *HudaqCtrReset;

  THudaqPWMWrite *HudaqPWMWrite;

  THudaqStepWrite *HudaqStepWrite;
} CallTable;

extern const CallTable CtDummy;
extern const CallTable CT612;     ///< AD612 resources
extern const CallTable CT614;     ///< MF614 resources
extern const CallTable CT622;     ///< AD622 resources
extern const CallTable CT624;     ///< MF624 resources


/** Obtain DeviceRecord from Hudaq handle. */
#define HudaqGetDeviceRecord(handle) ((DeviceRecord *)(handle))

int GenericDIReadBit(const DeviceRecord *DevRecord, unsigned channel, unsigned bit);
HUDAQSTATUS GenericAIReadMultiple(const DeviceRecord *DevRecord, unsigned number, const unsigned *channels, double *values);
HUDAQSTATUS GenericOneDIReadMultiple(const DeviceRecord *DevRecord, unsigned number, const unsigned* channels, unsigned* values);


#ifdef _MSC_VER
 #pragma warning(disable:4273) /* disable the "inconsistent DLL linkage" warning */
#endif //_MSC_VER


#if __GNUC__ == 3 && __GNUC_MINOR__ >= 3
 #ifdef HUDAQAPI
  #undef HUDAQAPI
  #define HUDAQAPI __attribute__ ((visibility("default")))
  #pragma visibility(hidden)
 #endif
#endif

int timeGetTime(void);

#define UNDEFINED_VALUE 0       ///< result of conversion when something goes bad

#endif
